/*
 * snoop.c
 * Copyright (C) 2005 Florin Malita <fmalita@gmail.com>
 *
 * This file is part of Snoop, a free GNU/Linux monitoring tool.
 * See http://snoop.sourceforge.net/ for details and updates.
 *
 * Snoop is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Snoop is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/namei.h>
#include <linux/mount.h>
#include <linux/sched.h>
#include <linux/file.h>
#include <linux/cdev.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <linux/wait.h>
#include <linux/ioctl.h>
#include <linux/device.h>
#include <linux/rcupdate.h>
#include <linux/poll.h>

#include <asm/atomic.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>

#ifdef HAVE_FDTABLE_H
#include <linux/fdtable.h>
#endif

#include "snoop.h"

MODULE_AUTHOR("Florin Malita <fmalita@gmail.com>");
MODULE_LICENSE("GPL");

static int snoop_major = 0;
static int snoop_max_open = DEFAULT_MAX_OPEN;
static int snoop_buffer_size = DEFAULT_BUFFER_SIZE;

static struct snoop_dev snoop_dev;

module_param(snoop_major, int, S_IRUGO);
module_param(snoop_max_open, int, S_IRUGO);
module_param(snoop_buffer_size, int, S_IRUGO);

static int attach(struct snoop *, struct file *filp, int mode);
static int detach(struct snoop_handle *, struct file *filp, int mode);

static inline void get_snoop(struct snoop *snoop)
{
	/* This might seem racy but the only way to get here 
	   is through snoop_open()/snoop_ioctl() and that means
	   we already have a refcount */
	__module_get(THIS_MODULE);
	atomic_inc(&snoop->count);
	
	TRACE("snoop (%s) usage count: %d\n", snoop->name, 
				atomic_read(&snoop->count));
}

static inline void put_snoop(struct snoop *snoop)
{
	if (atomic_dec_and_test(&snoop->count)){
		/* Can be reached only after snoop_release()
		   so there's no race with attach()/get_snoop() */
		TRACE("dropping snoop (%s)\n", snoop->name);
		kfree(snoop->buf);
		kfree(snoop);
	} else {
		TRACE("snoop (%s) usage count: %d\n", snoop->name, 
					atomic_read(&snoop->count));
	}
	
	module_put(THIS_MODULE);
}

static inline int buffer_space(struct snoop *snoop)
{
	int res;
	
	if (snoop->head == snoop->tail)
		res = snoop_buffer_size - 1;
	else
		res = ((snoop->tail + snoop_buffer_size - snoop->head)
				% snoop_buffer_size) - 1;
				
	TRACE("%d bytes free in buffer\n", res);
	
	return res;
}

static int data_in(struct snoop *snoop, char __user *buf, size_t count)
{
	size_t len;
	int pumped;
	
	TRACE("%d bytes available\n", count);

	if (down_interruptible(&snoop->sem))
		return -ERESTARTSYS; 
		/* FIXME: ^this^ isn't propagated all the way back - 
		   hopefully not a problem */

	while ((len = min(count, (size_t)buffer_space(snoop)))){

		if (snoop->head >= snoop->tail)
			len = min(len, (size_t)(snoop_buffer_size - snoop->head));
		else
			len = min(len, (size_t)(snoop->tail - snoop->head - 1));
	
		TRACE("attempting to pump %d bytes\n", len);
	
		pumped = len - copy_from_user(snoop->buf + snoop->head, buf, len);
	
		TRACE("managed to pump %d bytes\n", pumped);
	
		if (likely(pumped > 0)){
			snoop->head = (snoop->head + pumped) % snoop_buffer_size;
			count -= pumped;
			buf += pumped;
		} else {
			break;
		}
	}
	
	up(&snoop->sem);
	
	wake_up_interruptible(&snoop->inq);
	
	return 0;
}

static ssize_t read_hook(struct file *filp, char __user *buf, size_t count,
		loff_t *f_pos)
{
	struct snoop_handle *handle = container_of(filp->f_op, 
						struct snoop_handle, ops);
	struct snoop *snoop = handle->snoop;
	int res;
	
	TRACE("%d bytes on %s\n", count, handle->name);
	
	filp->f_op = handle->old_ops;

	if (handle->old_ops->read)
		res = handle->old_ops->read(filp, buf, count, f_pos);
	else
		res = do_sync_read(filp, buf, count, f_pos);
		
	handle->old_ops = filp->f_op;
	filp->f_op = &handle->ops;

	if (likely(res > 0))
		data_in(snoop, buf, res);
	
	return res;
}

static ssize_t write_hook(struct file *filp, const char __user *buf, size_t count,
		loff_t *f_pos)
{
	struct snoop_handle *handle = container_of(filp->f_op, 
						struct snoop_handle, ops);
	struct snoop *snoop = handle->snoop;
	int res;
	
	TRACE("%d bytes on %s\n", count, handle->name);

	data_in(snoop, (char __user *)buf, count);
	
	filp->f_op = handle->old_ops;

	if (handle->old_ops->write)
		res = handle->old_ops->write(filp, buf, count, f_pos);
	else
		res = do_sync_write(filp, buf, count, f_pos);

	handle->old_ops = filp->f_op;
	filp->f_op = &handle->ops;
	
	return res;
}

static int ioctl_hook(struct inode *inode, struct file *filp, unsigned int cmd,
			unsigned long arg)
{
	struct snoop_handle *handle = container_of(filp->f_op, 
						struct snoop_handle, ops);
	int res;

	TRACE("ioctl: 0x%u\n", cmd);
	
	filp->f_op = handle->old_ops;
	res = handle->old_ops->ioctl(inode, filp, cmd, arg);
	handle->old_ops = filp->f_op;
	filp->f_op = &handle->ops;

	return res;	
}

static int release_hook(struct inode *inode, struct file *filp)
{
	struct snoop_handle *handle = container_of(filp->f_op, 
						struct snoop_handle, ops);
	struct snoop *snoop = handle->snoop;
	int res;
	
	TRACE("...\n");
	
	if (down_interruptible(&snoop->sem))
		return -ERESTARTSYS;
	detach(handle, filp, 0);
	up(&snoop->sem);
	
	res = filp->f_op->release(inode, filp);
	
	put_snoop(snoop);
	
	return res;
}

static int snoop_open(struct inode *inode, struct file *filp)
{
	int res;
	struct snoop *snoop;
	
	TRACE("...\n");
	
	if (atomic_inc_return(&snoop_dev.open_count) >= snoop_max_open){
		ERROR("too many open snoop devices.\n");
		res = -ENFILE;
		goto out_count;
	}
	
	snoop = kmalloc(sizeof(struct snoop), GFP_KERNEL);
	if (!snoop){
		ERROR("kmalloc() failed.\n");
		res = -ENOMEM;
		goto out_count;
	}
	
	memset(snoop, 0, sizeof(struct snoop));
	
	snoop->buf = kmalloc(snoop_buffer_size, GFP_KERNEL);
	if (!snoop->buf){
		ERROR("kmalloc() failed.\n");
		res = -ENOMEM;
		goto out_snoop;
	}
	
	init_MUTEX(&snoop->sem);
	INIT_LIST_HEAD(&snoop->handles);
	init_waitqueue_head(&snoop->inq);
	snoop->head = snoop->tail = 0;
	atomic_set(&snoop->count, 0);
	
	filp->private_data = snoop;
	
	get_snoop(snoop);
	
	return 0;
out_snoop:
	kfree(snoop);
out_count:
	atomic_dec(&snoop_dev.open_count);
	return res;
}


static ssize_t snoop_read(struct file *filp, char __user *buf, size_t count,
		loff_t *f_pos)
{
	struct snoop *snoop = (struct snoop*)filp->private_data;
	
	TRACE("...\n");

	if (atomic_read(&snoop->count) < 2){
		TRACE("not attached.");
		return 0;
	}

	if (down_interruptible(&snoop->sem))
		return -ERESTARTSYS;
		
	while (snoop->head == snoop->tail){
		up(&snoop->sem);
		if (filp->f_flags & O_NONBLOCK)
			return -EAGAIN;
		
		TRACE("no data in buffer, going to sleep...\n");
		if (wait_event_interruptible(snoop->inq, 
					(snoop->head != snoop->tail)))
			return -ERESTARTSYS;
			
		if (down_interruptible(&snoop->sem))
			return -ERESTARTSYS;
	}
	
	TRACE("we have some data.\n");

	if (snoop->head > snoop->tail){
		count = min(count, (size_t)(snoop->head - snoop->tail));
	} else {
		count = min(count, (size_t)(snoop_buffer_size - snoop->tail));
	}
	
	if (copy_to_user(buf, snoop->buf + snoop->tail, count)){
		up(&snoop->sem);
		return -EFAULT;
	}
	
	TRACE("read %d bytes\n", count);
	
	snoop->tail = (snoop->tail + count) % snoop_buffer_size;
		
	up(&snoop->sem);
	
	return count;
}

static unsigned int snoop_poll(struct file *filp, poll_table *wait)
{
	struct snoop *snoop = (struct snoop*)filp->private_data;
	unsigned int mask = 0;
	
	TRACE("...\n");

	if (atomic_read(&snoop->count) < 2){
		TRACE("not attached.");
		return -EBADF;
	}

	if (down_interruptible(&snoop->sem))
		return -ERESTARTSYS;

	poll_wait(filp, &snoop->inq, wait);
	if (snoop->head != snoop->tail)
		mask |= POLLIN | POLLRDNORM;

	up(&snoop->sem);
	
	return mask;
}

static int get_name(char *buf, char *usr)
{
	int i;
	
	for (i = 0; i < SNOOP_NAME_SIZE; i++){
		if (get_user(buf[i], ((char*)usr + i)))
			return -EFAULT;
		if (!buf[i])
			break;
	}
	
	if (i == SNOOP_NAME_SIZE){
		ERROR("file name too long\n");
		return -EINVAL;
	}
	
	TRACE("got name: %s\n", buf);
	
	return 0;
}

static int is_attached(struct snoop *snoop, struct file *filp)
{
	struct snoop_handle *handle;
	
	list_for_each_entry(handle, &snoop->handles, list){
		if (handle->filp == filp){
			TRACE("filp 0x%p is attached by handle 0x%p\n", 
							filp, handle);
			return 1;
		}
	}
	
	TRACE("filp 0x%p is not attached\n", filp);
	
	return 0;
}

static int attach(struct snoop *snoop, struct file *filp, int mode)
{
	struct snoop_handle *handle;
	
	TRACE("filp: 0x%p\n", filp);
	
	if (is_attached(snoop, filp))
		return 0;
	
	handle = kmalloc(sizeof(struct snoop_handle), GFP_ATOMIC);
	if (!handle){
		ERROR("atomic kmalloc failed, running out of memory?\n");
		return 0;
	}
	
	memset(handle, 0, sizeof(struct snoop_handle));
	
	handle->snoop = snoop;
	handle->filp = filp;
	handle->old_ops = filp->f_op;
	strcpy(handle->name, snoop->name);
	
	memcpy(&handle->ops, filp->f_op, sizeof(struct file_operations));	
	handle->ops.ioctl = ioctl_hook;
	handle->ops.release = release_hook;
	if (mode & SNOOP_MODE_READ)
		handle->ops.read = read_hook;
	if (mode & SNOOP_MODE_WRITE)
		handle->ops.write = write_hook;
	
	list_add(&handle->list, &snoop->handles);
	get_snoop(snoop);

	wmb();
	filp->f_op = &handle->ops;
	
	return 1;
}

static int detach(struct snoop_handle *handle, struct file *filp, int mode)
{
	TRACE("...\n");
	
	list_del(&handle->list);
	filp->f_op = handle->old_ops;
	kfree(handle);
	
	return 1;
}

static int attach_fd(struct snoop *snoop, pid_t pid, unsigned fd, int mode)
{
	struct task_struct *task;
	struct files_struct *files;
	struct file *filp;
	int res = 0;
	
	TRACE("pid: %u, fd: %u\n", pid, fd);

	SNOOP_COMPAT_PID_LOCK(&tasklist_lock);
	
	task = find_task_by_pid(pid);

	if (task) {
		task_lock(task);
		
		files = task->files;
		if (files) {
			SNOOP_COMPAT_FDT_LOCK(&files->file_lock);
			
			filp = fcheck_files(files, fd);
			if (filp)
				res = attach(snoop, filp, mode);
			
			SNOOP_COMPAT_FDT_UNLOCK(&files->file_lock);
		}
		
		task_unlock(task);
	}

	SNOOP_COMPAT_PID_UNLOCK(&tasklist_lock);

	return res;
}

static int attach_path(struct snoop *snoop, char __user *name, int mode)
{
	int res;
	unsigned pid, fd;
	
	TRACE("attach mode: %d\n", mode);
			
	res = get_name(snoop->name, name);
	if (res)
		return res;
	
	res = sscanf(snoop->name, "/proc/%u/fd/%u", &pid, &fd);
	if (res == 2)
		return attach_fd(snoop, pid, fd, mode);
			
	return -ENOTTY;
}

static int snoop_ioctl(struct inode *inode, struct file *filp, 
				unsigned int cmd, unsigned long arg)
{
	struct snoop *snoop = (struct snoop*)filp->private_data;
	int res;
	
	TRACE("...\n");
	
	if (! capable(CAP_SYS_ADMIN))
		return -EPERM;
	
	if (down_interruptible(&snoop->sem))
		return -ERESTARTSYS;
	
	switch (cmd) {
	case SNOOP_IOCATTACHR:
		TRACE("ATTACHR\n");
		res = attach_path(snoop, (char*)arg, SNOOP_MODE_READ);
		break;
	
	case SNOOP_IOCATTACHW:
		TRACE("ATTACHW\n");
		res = attach_path(snoop, (char*)arg, SNOOP_MODE_WRITE);
		break;
		
	case SNOOP_IOCATTACHRW:
		TRACE("ATTACHRW\n");
		res = attach_path(snoop, (char*)arg, SNOOP_MODE_READ 
							| SNOOP_MODE_WRITE);
		break;

	case SNOOP_IOCGCOUNT:
		TRACE("GCOUNT\n");
		res = put_user(atomic_read(&snoop->count) - 1, 
							(unsigned long*)arg);
		break;
		
	default:
		res = -ENOTTY;
	}
	
	up(&snoop->sem);
	
	return res;
}

static int snoop_release(struct inode *inode, struct file *filp)
{
	struct snoop *snoop = (struct snoop*)filp->private_data;
	
	TRACE("...\n");

	put_snoop(snoop);

	atomic_dec(&snoop_dev.open_count);
	
	return 0;
}

static struct file_operations snoop_fops = {
	.owner 		= THIS_MODULE,
	.read		= snoop_read,
	.poll		= snoop_poll,
	.ioctl		= snoop_ioctl,
	.open		= snoop_open,
	.release	= snoop_release,
};

static int __init snoop_init(void)
{
	int res;
	dev_t dev;
	
	TRACE("...\n");

	atomic_set(&snoop_dev.open_count, 0);
	
	if (snoop_major){
		dev = MKDEV(snoop_major, SNOOP_MINOR);
		res = register_chrdev_region(dev, SNOOP_MINORS, "snoop");
	} else {
		res = alloc_chrdev_region(&dev, SNOOP_MINOR, SNOOP_MINORS, 
								"snoop");
	}
	
	if (res < 0) {
		ERROR("cannot get major %d\n", snoop_major);
		goto out;
	}
	
	snoop_dev.csnoop = SNOOP_COMPAT_CLASS_CREATE(THIS_MODULE, "snoop");
	if (IS_ERR(snoop_dev.csnoop)){
		ERROR("cannot create snoop class\n");
		goto out_region;
	}

	if (IS_ERR(SNOOP_COMPAT_DEVICE_CREATE(snoop_dev.csnoop, dev, NULL, "snoop"))){
		ERROR("cannot add snoop device.\n");
		goto out_class;
	}

	cdev_init(&snoop_dev.cdev, &snoop_fops);
	snoop_dev.cdev.owner = THIS_MODULE;
	snoop_dev.cdev.ops = &snoop_fops;
	
	res = cdev_add(&snoop_dev.cdev, dev, 1);
	if (res){
		ERROR("cdev_add() failed.\n");
		goto out_dev;
	}

	return 0;
	
out_dev:
	SNOOP_COMPAT_DEVICE_DESTROY(snoop_dev.csnoop, dev);
out_class:
	SNOOP_COMPAT_CLASS_DESTROY(snoop_dev.csnoop);
out_region:
	unregister_chrdev_region(dev, SNOOP_MINORS);
out:
	return res;
}

static void __exit snoop_exit(void)
{
	dev_t dev = snoop_dev.cdev.dev;
	
	TRACE("...\n");
	
	cdev_del(&snoop_dev.cdev);
	SNOOP_COMPAT_DEVICE_DESTROY(snoop_dev.csnoop, dev);
	SNOOP_COMPAT_CLASS_DESTROY(snoop_dev.csnoop);
	unregister_chrdev_region(dev, SNOOP_MINORS);
}

module_init(snoop_init);
module_exit(snoop_exit);
