/*
 * snoop.c
 * Copyright (C) 2005 Florin Malita <fmalita@gmail.com>
 *
 * This file is part of Snoop, a free GNU/Linux monitoring tool.
 * See http://snoop.sourceforge.net/ for details and updates.
 *
 * Snoop is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Snoop is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#   include "config.h"
#endif

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <dirent.h>

#define _GNU_SOURCE
#include <getopt.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/select.h>

#ifdef HAVE_INOTIFY_H
#   include <inotify.h>
#else
#   include "inotify.h"
#   include "inotify-syscalls.h"
#endif

#include "list.h"
#include "snoop.h"

#define BLAH(x...)	if (log_level) { printf(x); }

volatile int 		got_sigint;
struct list_head	wd_hash[SNOOP_HASH_SIZE];
int			log_level = 1;

char *short_opts = "hd:qrvw";
struct option long_opts[] = {
	{ "help",	0, NULL, 'h' },
	{ "device",	0, NULL, 'd' },
	{ "quiet",	0, NULL, 'q' },
	{ "read",	0, NULL, 'r' },
	{ "version",	0, NULL, 'v' },
	{ "write",	0, NULL, 'w' },
	{ NULL,		0, NULL, '\0'}
};

void sigint_handler(int sig)
{
	got_sigint = 1;
}

void usage(char *name, FILE *f)
{
	fprintf(f, "Usage: %s [OPTION...] FILE...\n", name);
	fprintf(f, "  OPTIONS:\n");
	fprintf(f, "    -h, --help         display this help message and exit.\n");
	fprintf(f, "    -d, --device=dev   use the specified device node (default: /dev/snoop)\n");
	fprintf(f, "    -q, --quiet        don't print informative messages.\n");
	fprintf(f, "    -r, --read         snoop the read operation.\n");
	fprintf(f, "    -v, --version      display version information.\n");
	fprintf(f, "    -w, --write        snoop the write operation (default).\n");
	fprintf(f, "\n");
}

int snoop_open(char *dev)
{
	int fd;
	
	fd = open(dev, O_RDWR);
	if (fd < 0){
		perror("open()");
	}
	
	return fd;
}

int snoop_attach(int snoop_fd, char *file, int mode)
{
	int res;
	int ctl;
	
	if ((mode & SNOOP_MODE_READ) && (mode & SNOOP_MODE_WRITE))
		ctl = SNOOP_IOCATTACHRW;
	else if (mode & SNOOP_MODE_READ)
		ctl = SNOOP_IOCATTACHR;
	else
		ctl = SNOOP_IOCATTACHW;
		
	res = ioctl(snoop_fd, ctl, file);

	return res;
}

int get_count(int fd)
{
	unsigned long count;
	
	if (ioctl(fd, SNOOP_IOCGCOUNT, &count) < 0){
		perror("ioctl()");
		return -1;
	}
	
	return count;
}

int get_major()
{
	FILE *f;
	int maj = 0;
	char buf[128];

	if ((f = fopen("/proc/devices", "r")) == NULL){
		perror("fopen()");
		return 0;
	}

	while (fgets(buf, 128, f)){
		if (strstr(buf, MODULE_NAME)){
			sscanf(buf, "%d ", &maj);
			break;
		}
	}
	fclose(f);

	return maj;
}

int setup_dev(char *name)
{
	struct stat st;
	int maj, res;

	if (! (maj = get_major())) {
		system("modprobe " MODULE_NAME " &>/dev/null");

		/* give the module some time to initialize */
		usleep(100000);
		maj = get_major();
	}

	if (!maj)
		return -1;

	res = stat(name, &st);
	if ((res < 0) || (major(st.st_rdev) != maj)){
		unlink(name);
		res = mknod(name, S_IFCHR | S_IRUSR | S_IWUSR, makedev(maj, 0));
		if (res < 0){
			perror("mknod()");
			return -1;
		}
	}

	return 0;
}

static inline unsigned hash(int i)
{
	return i % SNOOP_HASH_SIZE;
}

static inline void hash_init()
{
	int i;
	
	for (i = 0; i < SNOOP_HASH_SIZE; i++)
		INIT_LIST_HEAD(&wd_hash[i]);
}

static inline struct snoop_wd* hash_add(int wd, char *name)
{
	struct snoop_wd *swd;
	int slen = strlen(name) + 1;
	
	if (!(swd = malloc(sizeof(struct snoop_wd)))){
		perror("malloc()");
		goto out_err;
	}
	
	if (!(swd->name = malloc(slen))) {
		perror("malloc()");
		goto out_swd;
	}
	
	swd->wd = wd;
	strcpy(swd->name, name);
	list_add_tail(&swd->list, &wd_hash[hash(wd)]);
	
	return swd;
	
out_swd:
	free(swd);
out_err:
	return NULL;
}

static inline void hash_del_swd(struct snoop_wd *swd)
{
	list_del(&swd->list);
	free(swd->name);
	free(swd);
}

static inline void hash_del_list(struct list_head *l)
{
	struct snoop_wd *swd, *t;
	
	list_for_each_entry_safe(swd, t, l, list)
		hash_del_swd(swd);
}

static inline void hash_destroy()
{
	int i;
	
	for (i = 0; i < SNOOP_HASH_SIZE; i++)
		hash_del_list(&wd_hash[i]);
}

static inline struct snoop_wd* hash_find(unsigned wd)
{
	struct snoop_wd *swd;
	
	list_for_each_entry(swd, &wd_hash[hash(wd)], list)
		if (swd->wd == wd)
			return swd;
			
	return NULL;

}

int is_num(const char *str)
{
	const char *c;

	for (c = str; *c; c++)
		if (!isdigit(*c))
			return 0;
	return 1;
}

void search_open(int snoop_fd, const char *name, int mode)
{
	char buf[128];
	char fdname[PATH_MAX];
	DIR *proc_dir, *fd_dir;
	struct dirent *proc_ent, *fd_ent;
	int res;

	if (!(proc_dir = opendir(PROC_DIR))) {
		perror("opendir()");
		return;
	}

	while ((proc_ent = readdir(proc_dir))) {
		if (! is_num(proc_ent->d_name))
			continue;
		
		if (snprintf(buf, 128, "%s/%s/fd", PROC_DIR,
			proc_ent->d_name) >= 128)
			continue;

		if (!(fd_dir = opendir(buf)))
			continue;

		while ((fd_ent = readdir(fd_dir))) {
			if (!is_num(fd_ent->d_name))
				continue;

			if (snprintf(buf, 128, "%s/%s/fd/%s", PROC_DIR,
				proc_ent->d_name, fd_ent->d_name) >= 128)
				continue;

			if ((res = readlink(buf, fdname, PATH_MAX)) < 0) {
				perror("readlink()");
				continue;
			}
			if (res >= PATH_MAX)
				continue;
			fdname[res] = '\0';

			if ((strcmp(name, fdname) == 0) &&
				(snoop_attach(snoop_fd, buf, mode) > 0))
				BLAH("[attached to %s (pid %s, fd %s), %d total fd(s)]\n",
					name, proc_ent->d_name, fd_ent->d_name,
					get_count(snoop_fd));
		}
		
		closedir(fd_dir);
	}

	closedir(proc_dir);
}

void handle_inotify_events(char *buf, int count, int fd, int ifd, int mode) {
	char *p;
	unsigned wd;
	struct inotify_event *iev;
	struct snoop_wd *swd;

	for (p = buf; p < buf + count; p += sizeof(struct inotify_event) + 
				((struct inotify_event*)p)->len) {
		iev = (struct inotify_event*)p;
		
		if (! (swd = hash_find(iev->wd)))
			continue;

		if ((iev->mask & IN_OPEN)) {
			search_open(fd, swd->name, mode);
		}

		if ((iev->mask & IN_CREATE && (iev->len))) {
			if (strcmp(iev->name, strrchr(swd->name, '/') + 1))
				continue;

			wd = inotify_add_watch(ifd, swd->name, IN_OPEN | 
				IN_DELETE_SELF);
			if (wd >= 0)
				hash_add(wd, swd->name);
			search_open(fd, swd->name, mode);
		}

		if ((iev->mask & IN_DELETE_SELF)) {
			inotify_rm_watch(ifd, iev->wd);
			hash_del_swd(swd);
		}
	}
}

void io_loop(int in, int out, int ifd, int mode)
{
	char buf[BUFFER_SIZE];
	char ibuf[BUFFER_SIZE];
	int head = 0, tail = 0;
	int res, buf_free, count;
	fd_set rd_set, wr_set;
	int max_fd, att_count;
	struct sigaction sa = {
		.sa_handler = sigint_handler,
	};
	
	BLAH("[CTRL-C to quit]\n");
	sigaction(SIGINT, &sa, NULL);
	
	max_fd = (in > out) ? in : out;
	max_fd = (ifd > max_fd) ? ifd : max_fd;
	
	buf_free = BUFFER_SIZE - 1;
	att_count = get_count(in);
	
	while (!got_sigint){
		FD_ZERO(&rd_set);
		FD_ZERO(&wr_set);
		
		if (ifd > 0)
			FD_SET(ifd, &rd_set);
		
		if (att_count && buf_free)
			FD_SET(in, &rd_set);
		if (head != tail)
			FD_SET(out, &wr_set);
		
		res = select(max_fd + 1, &rd_set, &wr_set, NULL, NULL);
		if (res < 0){
			if (errno != EINTR){
				perror("select()");
				break;
			}
			continue;
		}
		
		if (FD_ISSET(in, &rd_set)) {
			count = (head >= tail) ? (BUFFER_SIZE - head) : \
						(tail - head -1 );
			count = (count > buf_free) ? buf_free : count;
			
			if ((res = read(in, buf + head, count)) < 0) {
				perror("read()");
				break;
			}
			
			if (res == 0)
				att_count--;
			
			head = (head + res) % BUFFER_SIZE;
			buf_free -= res;
		}
		
		if (FD_ISSET(out, &wr_set)) {
			count = (head >= tail) ? (head - tail) : \
						(BUFFER_SIZE - tail);
						
			res = write(out, buf + tail, count);
			if (res < 0) {
				perror("write()");
				break;
			}
			
			tail = (tail + res) % BUFFER_SIZE;
			buf_free += res;
		}
		
		if ((ifd > 0) && FD_ISSET(ifd, &rd_set)) {
			res = read(ifd, ibuf, BUFFER_SIZE);
			if (res <= 0) {
				perror("read()");
				break;
			}
			
			handle_inotify_events(ibuf, res, in, ifd, mode);
			att_count = get_count(in);
		}
	}
	BLAH("\n[snoop detached]\n");
}

int main(int argc, char **argv)
{
	signed char opt;
	char *c, *snoop_dev = DEFAULT_SNOOP_DEV;
	int mode = 0;
	int fd, ifd, wd, res;
	
	hash_init();
	
	while ((opt = getopt_long(argc, argv, short_opts, long_opts, NULL)) >= 0){
		switch (opt){
		case 'h':
			usage(argv[0], stdout);
			exit(0);
			
		case 'd':
			snoop_dev = optarg;
			break;
			
		case 'q':
			log_level = 0;
			break;
			
		case 'r':
			mode |= SNOOP_MODE_READ;
			break;
		
		case 'v':
			fprintf(stderr, "%s version %s.\n", argv[0], VERSION);
			exit(0);
		
		case 'w':
			mode |= SNOOP_MODE_WRITE;
			break;
			
		case '?':
		default:
			usage(argv[0], stderr);
			exit(1);
		}
	
	}

	if (!mode)
		mode = SNOOP_MODE_WRITE;

	if (setup_dev(snoop_dev) < 0){
		fprintf(stderr, "could not set up %s\n", snoop_dev);
		return 1;
	}

	if ((fd = snoop_open(snoop_dev)) < 0)
		return 1;

	if (optind == argc){
		usage(argv[0], stderr);
		return 1;
	}

	if ((ifd = inotify_init()) < 0)
		perror("inotify_init()");

	for (; optind < argc; optind++) {
		if (sscanf(argv[optind], PROC_DIR "/%u/" FD_DIR "/%u",
			&res, &res) == 2) {
			snoop_attach(fd, argv[optind], mode);
		} else {
			/* watch the file and its parent dir */
			if (ifd > 0) {
				wd = inotify_add_watch(ifd, argv[optind],
					IN_OPEN | IN_DELETE_SELF);
				if (wd >= 0)
					hash_add(wd, argv[optind]);

				if ((c = strrchr(argv[optind], '/')) &&
					(c != argv[optind])) {
					
					*c = '\0';
					wd = inotify_add_watch(ifd, argv[optind],
						IN_CREATE);
					*c = '/';
					if (wd >= 0)
						hash_add(wd, argv[optind]);
				}
			}
			search_open(fd, argv[optind], mode);
		}
	}

	if (get_count(fd) >= 0)
		io_loop(fd, 1, ifd, mode);

	hash_destroy();
	
	close(ifd);
	close(fd);

	return 0;
}
